/* interp.c

	reentrant eval for som3j		22 Dec 2004
	som-in-som						30 Dec 2004
	add icRef for Rz33				20 Aug 2011
 */

// system area  1 .. MAXSYS-1

#include "compile.h"
#include "interp.h"

#define		PRIVATE 	static

int M[MAXMEM];	 				// memory
int *CS = &M[MAXDS];			// code segment
PRIVATE int *SS = &M[MAXCS];	// stack segment, merge to M[.]
//PRIVATE int SS[MAXSS];		// stack segment
PRIVATE	int sp=1,fp=0;			// current state SS

int DP = MAXSYS;		// current free data segment

void error(char *mess, int ip){
	printf("%s at ip %d\n",mess,ip);
	exit(0);
}

// static alloc memory of size a
int xalloc(int a){
	int b;
	b = DP;
	if( DP + a >= MAXDS )
		error("xalloc: out of memory",0);
	DP += a;
	return b;
}

// gets() and put it in M[ads]
void dogets(int ads){
	char buf[256], *c;
	int i;
	c = gets(buf);
	if( c != NULL ){
		for(i=0; buf[i] != 0; i++, ads++)
			M[ads] = (int)buf[i] & 255;
		M[ads] = 0;
	}else{
		M[ads] = EOF_CHAR;
		M[ads+1] = 0;
	}
}

// reentrant eval
void eval(int ref){
	int runflag, a;
	int ip, op, data;	// current instruction
	int cnt;

	cnt = 0;
	ip = ref;
	runflag = 1;
    while(runflag ) {

		if(cnt > 2000000) {printf("infinite loop\n"); break;}
		cnt++;

		op = CS[ip] & 255;
		data = CS[ip] >> 8 ;

	switch(op){
	case icLit:
		push(data);
		ip++;
		break;
	case icLd:
		push( M[data]);
		ip++;
		break;
	case icSt:
		M[data] = TOS;
		sp--;
		ip++;
		break;
	case icPut:
		SS[fp-data] = TOS;
		sp--;
		ip++;
		break;
	case icGet:
		push(SS[fp-data]);
		ip++;
		break;
	case icLdx:
		NEXT = M[NEXT + TOS];
		sp--;
		ip++;
		break;
	case icStx:
		pop(a);
		M[NEXT + TOS] = a;
		sp -= 2;
		ip++;
		break;
	case icArray:
		pop(a);
		push(xalloc(a));
		ip++;
		break;
	case icJmp:
		ip += data;
		break;
	case icJf:
		pop(a);
		if( a == FALSE) ip += data;
		else ip++;
		break;
	case icJt:
		pop(a);
		if( a != FALSE) ip += data;
		else ip++;
		break;
	case icCall:
		a = CS[data] >> 8;		// get lv
		if (sp + a + FUDGE > N_SS)
			error("stack overflow",ip);
		SS[sp+a] = fp;
		fp = sp + a;
		sp = fp + 1;
		SS[sp] = ip + 1;
		ip = data + 1;
		break;
	case icRet:
	// return TOS if current stack is not empty
		ip = SS[fp+1];
		if ( sp > fp+1 ){
			a = TOS;
			sp = fp - data + 1;
			fp = SS[fp];
			SS[sp] = a;
		}else{
			sp = fp - data;
			fp = SS[fp];
		}
		break;
	case icAdd:
		NEXT += TOS;
		sp--;
		ip++;
		break;
	case icSub:
		NEXT -= TOS;
		sp--;
		ip++;
		break;
	case icMul:
		NEXT *= TOS;
		sp--;
		ip++;
		break;
	case icDiv:
		NEXT /= TOS;
		sp--;
		ip++;
		break;
	case icBand:
		NEXT &= TOS;
		sp--;
		ip++;
		break;
	case icBor:
		NEXT |= TOS;
		sp--;
		ip++;
		break;
	case icBxor:
		NEXT ^= TOS;
		sp--;
		ip++;
		break;
	case icNot:
		TOS = TOS == FALSE;
		ip++;
		break;
	case icMod:
		NEXT %= TOS;
		sp--;
		ip++;
		break;
	case icShl:
		NEXT <<= TOS;
		sp--;
		ip++;
		break;
	case icShr:
		NEXT >>= TOS;
		sp--;
		ip++;
		break;
	case icEq:
		NEXT = NEXT == TOS;
		sp--;
		ip++;
		break;
	case icNe:
		NEXT = NEXT != TOS;
		sp--;
		ip++;
		break;
	case icLt:
		NEXT = NEXT < TOS;
		sp--;
		ip++;
		break;
	case icLe:
		NEXT = NEXT <= TOS;
		sp--;
		ip++;
		break;
	case icGt:
		NEXT = NEXT > TOS;
		sp--;
		ip++;
		break;
	case icGe:
		NEXT = NEXT >= TOS;
		sp--;
		ip++;
		break;
	case icInc:
		SS[fp-data]++;
		ip++;
		break;
	case icDec:
		SS[fp-data]--;
		ip++;
		break;
/*
	case icCase:
		low = CS[ip+1] >> 8;
		hi = CS[ip+2] >> 8;
		pop(a);
		if( a < low || a > hi )
			ip = ip+3;
		else
			ip += a-low+4;
		break;
*/
	case icRef:					// &lv
		push(MAXCS+fp-data);	// absolute address
		ip++;
		break;
	case icSys:
		switch(data){
		case 1: 			// print int
			pop(a);
			printf("%d",a);
			break;
		case 2: 			// print char
			pop(a);
			printf("%c",a);
			break;
		case 3: 			// get char
			push(getchar());
			break;
		case 4:			   // gets()
		    pop(a);		   // som buffer
			dogets(a);
			break;
/*
		case 5:				// eval ref
			pop(a);
			CS = &M[CS_SOM]; // switch to user space
			eval(a);
			CS = &M[MAXDS];	// switch back
			break;
*/
		case 6:				// print array-string in M[]
			pop(a);			//    new for rz33   20 Aug 2011
			while(M[a] != 0){
				if(M[a] == 92 && M[a+1] == 110){	// newline
					printf("\n");
					a++;
				}else
					printf("%c",M[a]);
				a++;
			}
			break;
		case 7:
			runflag = 0;	 	// end
			break;

		default: error("undefined syscall",ip);
		}
		ip++;
		break;

	case icEnd:
		runflag = 0;
		break;
	default:
		printf("op %d\n",op);
		error("undefined opcode",ip);
	}
	// end swich(op)

    } // end while runflag
}


